import json
import time

from golem.core.chat_session import ChatSession
from golem.core.logging.abs_logger import MessageLogger


def get_elastic():
    from elasticsearch import Elasticsearch
    from django.conf import settings
    config = settings.GOLEM_CONFIG.get('ELASTIC')
    if not config:
        return None
    return Elasticsearch(config['HOST'], port=config['PORT'])


class ElasticsearchLogger(MessageLogger):
    def __init__(self):
        super().__init__()
        self.test_id = -1  # FIXME

    def log_user_message(self, dialog, time, state, message, type_, entities):

        entities = entities.copy()
        text = None
        if '_message_text' in entities:
            text = entities['_message_text'][0]['value']
            del (entities['_message_text'])

        message = {
            'uid': dialog.session.chat_id,
            'test_id': self.test_id,
            'created': time,
            'is_user': True,
            'text': text,
            'state': state,
            'type': type_,
            'entities': entities
        }
        self._log_message(message)

    def log_bot_message(self, dialog, time, state, message):
        from golem.core.responses import TextMessage
        type_ = type(message).__name__ if message else 'TextMessage'
        response = json.loads(
            json.dumps(message, default=lambda obj: obj.__dict__ if hasattr(obj, '__dict__') else str(obj)))

        text = message.text if hasattr(message, 'text') else str(message)

        message = {
            'uid': dialog.session.chat_id,
            'test_id': self.test_id,
            'created': time,
            'is_user': False,
            'text': text,
            'state': state,
            'type': type_,
            'response': response,
        }
        self._log_message(message)

    def log_error(self, dialog, state, exception):
        message = {
            'uid': dialog.chat_id,
            'test_id' : self.test_id,
            'created': time.time(),
            'is_user': False,
            'text': str(exception),
            'state': state,
            'type': 'error'
        }
        self._log_message(message)

    def _log_message(self, message):
        es = get_elastic()
        if not es:
            return
        try:
            print('Logging', message)
            es.index(index="message-log", doc_type='message', body=message)
        except Exception as e:
            print('Unable to log message to Elasticsearch.')
            print(e)

    def log_user(self, dialog, session: ChatSession):
        user = {
            'uid': session.chat_id,
            'profile': session.profile.to_json() if session.profile else None
        }
        es = get_elastic()
        if not es:
            return
        try:
            es.create(index="message-log", id=user['uid'], doc_type='user', body=user)
        except Exception as e:
            print('Unable to log user profile to Elasticsearch.')
            print(e)



import datetime
import json
import logging
import time
import traceback

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http.response import HttpResponse, JsonResponse
from django.shortcuts import render
from django.template import loader
from django.utils.decorators import method_decorator
from django.views import generic
from django.views.decorators.csrf import csrf_exempt

from golem.core.interfaces.facebook import FacebookInterface
from golem.core.interfaces.microsoft import MicrosoftInterface
from golem.core.interfaces.telegram import TelegramInterface
from golem.core.persistence import get_redis
from golem.core.logging.elastic import get_elastic
from golem.core.tests import ConversationTest, ConversationTestRecorder, ConversationTestException, TestLog, \
    UserTextMessage


class FacebookView(generic.View):

    def get(self, request, *args, **kwargs):
        if self.request.GET.get('hub.verify_token') == settings.GOLEM_CONFIG.get('WEBHOOK_VERIFY_TOKEN'):
            return HttpResponse(self.request.GET['hub.challenge'])
        else:
            return HttpResponse('Error, invalid token')

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return generic.View.dispatch(self, request, *args, **kwargs)

    # Post function to handle Facebook messages
    def post(self, request, *args, **kwargs):
        # Converts the text payload into a python dictionary
        request_body = json.loads(self.request.body.decode('utf-8'))
        FacebookInterface.accept_request(request_body)
        return HttpResponse()


class TelegramView(generic.View):

    def get(self, request, *args, **kwargs):
        request_body = json.loads(self.request.body.decode('utf-8'))
        from pprint import pprint
        pprint(request_body)
        return HttpResponse()

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return generic.View.dispatch(self, request, *args, **kwargs)

    # Post function to handle Telegram messages
    def post(self, request, *args, **kwargs):
        # Converts the text payload into a python dictionary
        request_body = json.loads(self.request.body.decode('utf-8'))
        TelegramInterface.accept_request(request_body)
        return HttpResponse()


class GActionsView(generic.View):
    def get(self, request, *args, **kwargs):
        return HttpResponse()

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return generic.View.dispatch(self, request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        from golem.core.interfaces.google import GoogleActionsInterface
        body = json.loads(self.request.body.decode('utf-8'))
        logging.info(body)
        resp = GoogleActionsInterface.accept_request(body)
        return JsonResponse(resp)


class SkypeView(generic.View):
    def get(self, request, *args, **kwargs):
        body = json.loads(self.request.body.decode('utf-8'))
        MicrosoftInterface.accept_request(body)
        return HttpResponse()

    def post(self, request, *args, **kwargs):
        body = json.loads(self.request.body.decode('utf-8'))
        MicrosoftInterface.accept_request(body)
        return HttpResponse()

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        return generic.View.dispatch(self, request, *args, **kwargs)

@login_required
def run_all_tests(request):
    modules = _get_test_modules('./tests/')

    print('Running tests: {}'.format(modules))
    tests = []
    db = get_redis()
    db.delete('test_results')
    for module in modules:
        print('Running tests "{}"'.format(module))
        tests.append(_run_test_module(module))

    db.set('test_time', str(datetime.datetime.now()))

    return JsonResponse(data=tests, safe=False)

def _get_test_modules(path):
    from os import listdir
    from os.path import join
    return sorted([f.replace('.py','') for f in listdir(path) if join(path, f).endswith('.py') and not f.startswith('_')])

@login_required
def test(request):
    db = get_redis()
    results = db.hgetall('test_results')
    results = [json.loads(results[k].decode('utf-8')) for k in sorted(list(results))] if results else []

    updated_time = db.get('test_time')
    updated_time = db.get('test_time').decode('utf-8') if updated_time else None

    status = 'passed'
    avg = {'duration':0, 'total':0, 'init':0, 'parsing':0, 'processing':0}
    passed = 0
    for test in results:
        result = test['result']
        if result['status'] == 'passed':
            passed += 1
            avg['duration'] += result['duration']
            avg['total'] += result['report']['avg']['total']
            avg['init'] += result['report']['avg']['init']
            avg['parsing'] += result['report']['avg']['parsing']
            avg['processing'] += result['report']['avg']['processing']
        elif status != 'exception':
            status = result['status']

    if passed > 0:
        for key in avg:
            avg[key] = avg[key] / passed

    context = {'tests':results, 'avg':avg, 'status':status, 'updated_time' : updated_time}
    template = loader.get_template('golem/test.html')
    return HttpResponse(template.render(context, request))

def run_test(request, name):
    benchmark = request.GET.get('benchmark', False)
    return JsonResponse(data=_run_test_module(name, benchmark=benchmark))

def run_test_message(request, message):
    return _run_test_actions('message', [UserTextMessage(message)])

def _run_test_module(name, benchmark=False):
    import importlib

    module = importlib.import_module('tests.'+name)
    importlib.reload(module)

    result = _run_test_actions(name, module.actions, benchmark=benchmark)
    test = {'name':name, 'result':result}
    db = get_redis()
    db.hset('test_results', name, json.dumps(test))
    return test

def _run_test_actions(name, actions, benchmark=False):
    test = ConversationTest(name, actions, benchmark=benchmark)
    start_time = time.time()
    report = None
    try:
      report = test.run()
    except Exception as e:
      log = TestLog.get()
      fatal = not isinstance(e, ConversationTestException)
      if fatal:
        trace = traceback.format_exc()
        print(trace)
        log.append(trace)
      return {'status': 'exception' if fatal else 'failed', 'log':log, 'message':str(e), 'report':report}

    elapsed_time = time.time() - start_time
    return {'status': 'passed', 'log':TestLog.get(), 'duration':elapsed_time, 'report':report}

def test_record(request):
    response = ConversationTestRecorder.get_result()

    response = HttpResponse(content_type='application/force-download', content=response) #
    response['Content-Disposition'] = 'attachment; filename=mytest.py'
    return response

@login_required
def log_tests(request):

    es = get_elastic()
    if not es:
        return HttpResponse('not able to connect to elasticsearch')

    res = es.search(index="message-log", doc_type='message', body={
    "size": 0,
    "aggs" : {
        "test_ids" : {
            "terms" : { "field" : "test_id",  "size" : 500 }
        }
    }})
    test_ids = []
    for bucket in res['aggregations']['test_ids']['buckets']:
        test_id = bucket['key']
        test_name = test_id.replace('_',' ').capitalize()
        test_ids.append({'id':'test_id_'+test_id, 'name':test_name})


    context = {
        'groups' : test_ids
    }
    template = loader.get_template('golem/log.html')
    return HttpResponse(template.render(context,request))

@login_required
def log(request, user_limit):

    user_limit = int(user_limit) if user_limit else 100
    es = get_elastic()
    if not es:
        return HttpResponse('not able to connect to elasticsearch')

    res = es.search(index="message-log", doc_type='message', body={
       "size": 0,
       "aggs": {
          "uids": {
             "terms": {
                "field": "uid",
                "size": 1000
             },
             "aggs": {
                "created": {
                   "max": {
                      "field": "created"
                   }
                }
             }
          }
       }
    })
    uids = []
    for bucket in res['aggregations']['uids']['buckets']:
        uid = bucket['key']
        last_time = bucket['created']['value']
        uids.append({'uid':uid, 'last_time':last_time})

    uids = sorted(uids, key=lambda uid: -uid['last_time'])[:user_limit]


    res = es.search(index="message-log", doc_type='user', body={
       "size" : user_limit,
       "query": {
            "bool" : {
                "filter" : {
                    "terms" : { "uid" : [u['uid'] for u in uids] }
                }
            }
        }
    })

    user_map = {user['_source']['uid']: {'name' : '{} {}'.format(user['_source']['profile'].get('first_name'), user['_source']['profile'].get('last_name')), 'id' : 'uid_'+user['_source']['uid']} for user in res['hits']['hits'] if user['_source']}
    users = [user_map[u['uid']] if u['uid'] in user_map else {'id':'uid_'+u['uid'], 'name':u['uid']} for u in uids]

    context = {
        'groups': users
    }

    print(users)
    template = loader.get_template('golem/log.html')
    return HttpResponse(template.render(context,request))

def log_conversation(request, group_id=None, page=1):
    page = int(page) if page else 1
    es = get_elastic()
    if not es:
        return HttpResponse()

    term = {}

    if group_id.startswith('test_id'):
        term = {"test_id" : group_id.replace('test_id_','')}
    else:
        term = { "uid" : group_id.replace('uid_','')}

    res = es.search(index="message-log", doc_type='message', body={
       "size": 50,
       "from" : 50*(page-1),
       "query": {
            "bool" : {
                "filter" : {
                    "term" : term
                }
            }
        },
       "sort": {
          "created": {
             "order": "desc"
          }
       }
    })

    messages = []
    previous = None
    for hit in res['hits']['hits'][::-1]:
        message = hit['_source']
        message['switch'] = previous != message['is_user']
        previous = message['is_user']
        response = message.get('response')
        elements = response.get('elements') if response else None
        if elements:
            message['elementWidth'] = len(elements)*215;
        message['json'] = json.dumps(message)
        messages.append(message)

    context = {'messages': messages}
    template = loader.get_template('golem/log_conversation.html')
    return HttpResponse(template.render(context,request))

def debug(request):
    FacebookInterface.accept_request({'entry':[{'messaging':[{'message': {'seq': 356950, 'mid': 'mid.$cAAPhQrFuNkFibcXMZ1cPICEB8YUn', 'text': 'hi'}, 'recipient': {'id': '1092102107505462'}, 'timestamp': 1595663674471, 'sender': {'id': '1046728978756975'}}]}]})

    return HttpResponse('done')


def users_view(request):
    from golem.models import User
    offset = int(request.GET.get("offset", 0))
    limit = int(request.GET.get("limit", 50))
    users = User.objects.all().order_by("uid")[offset:limit + offset]
    context = {"users": users, "next_offset": offset + limit, "prev_offset": offset - limit}
    return render(request, "golem/users.html", context)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-12-26 09:54
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='User',
            fields=[
                ('uid', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('first_name', models.CharField(blank=True, max_length=255, null=True)),
                ('surname', models.CharField(blank=True, max_length=255, null=True)),
                ('last_active', models.DateTimeField(blank=True, null=True)),
                ('message_cnt', models.PositiveIntegerField(default=0)),
            ],
        ),
    ]



import logging

import os
import yaml

from golem.nlu import utils


class GolemNLU:

    def __init__(self):
        self.nlp_data_dir = utils.data_dir()
        self.model_dir = os.path.join(self.nlp_data_dir, 'model')

        self.entities = {}

        entity_names = [i for i in os.scandir(self.model_dir) if i.is_dir()]
        for dir in entity_names:
            self.load_entity(dir.name)

    def load_entity(self, name):

        base_dir = os.path.join(self.model_dir, name)

        with open(os.path.join(base_dir, 'metadata.json'), 'r') as f:
            metadata = yaml.load(f)

        strategy = metadata['strategy']

        model = None

        if strategy == 'bow':
            from golem.nlu.nn.bow_model import BowModel
            model = BowModel(name, base_dir)
        elif metadata['strategy'] == 'context':
            from golem.nlu.nn.contextual import ContextualModel
            model = ContextualModel(name, base_dir)
        elif metadata['strategy'] == 'trait':
            from golem.nlu.nn.rnn_intent_model import RecurrentIntentModel
            model = RecurrentIntentModel(name, base_dir)
        elif metadata['strategy'] == 'keywords':
            from golem.nlu.keywords import TrieKeywordModel
            model = TrieKeywordModel(name, base_dir)
        else:
            logging.warning(
                'Unknown search strategy {} for entity {}, not training!'
                    .format(metadata['strategy'], name)
            )

        if model is not None:
            self.entities[name] = model

        logging.info("Loaded entity {name}".format(name=name))

    def parse(self, utterance):

        output = {}

        for entity in self.entities:

            model = self.entities[entity]
            result = model.predict(utterance)

            if result is None:
                continue
            elif isinstance(result, list):
                output.setdefault(entity, []).extend(result)
            elif isinstance(result, dict):
                output.setdefault(entity, []).append(result)
            else:
                logging.error("Invalid NLU response of type {}".format(type(result)))

        return output

    def parse_entity(self, utterance, entity, threshold=None):
        model = self.entities.get(entity, None)
        if model is not None:
            if threshold:
                return model.predict(utterance, threshold)
            else:
                return model.predict(utterance)
        return None



import json
import logging
import string
from datetime import datetime, timedelta
from typing import Optional

import random
import requests
from django.conf import settings

from golem.core.message_parser import parse_text_message
from golem.core.persistence import get_redis
from golem.tasks import accept_user_message


# FIXME needs to be updated

class TelegramInterface:
    name = 'telegram'
    prefix = 'tg'

    @staticmethod
    def get_base_url() -> Optional[str]:
        token = settings.GOLEM_CONFIG.get('TELEGRAM_TOKEN')
        if not token:
            logging.warning('Telegram token not provided. Telegram will not work.')
            return None
        return 'https://api.telegram.org/bot{}/'.format(token)

    @staticmethod
    def init_webhooks():
        from django.urls import reverse

        base_url = TelegramInterface.get_base_url()
        if not base_url:
            return
        url = base_url + 'setWebhook'

        print('Reverse telegram is: {}'.format(reverse('telegram')))
        callback_url = settings.GOLEM_CONFIG.get('DEPLOY_URL') + reverse('telegram')

        payload = {'url': callback_url}
        response = requests.post(url, data=payload)
        if not response.json()['ok']:
            logging.warning(response.json())


    @staticmethod
    def clear():
        pass

    @staticmethod
    def load_profile(uid):
        return {'first_name': 'Tests', 'last_name': ''}

    @staticmethod
    def post_message(uid, chat_id, response):
        from golem.core.interfaces.adapter.telegram import TelegramAdapter
        base_url = TelegramInterface.get_base_url()
        if not base_url:
            return
        adapter = TelegramAdapter(chat_id)
        messages = adapter.to_response(response)
        for method, payload in messages:
            url = base_url + method
            response = requests.post(url, data=payload)
            if not response.json()['ok']:
                logging.error('Telegram request failed!')
                logging.error(response.json())
                logging.error('for method {}'.format(method))
                logging.error('message is:')
                logging.error(payload)
                return

    @staticmethod
    def answer_callback_query(query_id, chat_id, message_id):
        base_url = TelegramInterface.get_base_url()
        if not base_url:
            return
        url = base_url + 'answerCallbackQuery'
        payload = {
            'callback_query_id': str(query_id)
        }
        response = requests.post(url, data=payload)
        if not response.json()['ok']:
            logging.error('Unable to answer callback query')
            logging.error(response.json())
        # hide reply keyboard after clicking
        url = base_url + 'editMessageReplyMarkup'
        payload = {'chat_id': chat_id, 'message_id': message_id, 'reply_markup': ''}
        response = requests.post(url, data=payload)
        if not response.json()['ok']:
            logging.error('Unable to remove quick replies')
            logging.error(response.json())

    @staticmethod
    def send_settings(settings):
        pass

    @staticmethod
    def processing_start(uid, chat_id):
        base_url = TelegramInterface.get_base_url()
        if not base_url:
            return
        url = base_url + 'sendChatAction'
        payload = {
            'chat_id': chat_id,
            'action': 'typing'
        }
        response = requests.post(url, data=payload)
        if not response.json()['ok']:
            logging.warning(response.json())

    @staticmethod
    def processing_end(uid, chat_id):
        pass

    @staticmethod
    def state_change(state):
        pass

    @staticmethod
    def has_message_expired(message: dict) -> bool:
        if not (message and 'date' in message):
            logging.warning('Invalid date in message')
            return True
        received = datetime.fromtimestamp(int(message['date']))
        now = datetime.utcnow()

        if abs(now - received) > timedelta(seconds=settings.GOLEM_CONFIG['MSG_LIMIT_SECONDS']):
            logging.warning('Ignoring message, too old')
            return True
        return False

    @staticmethod
    def accept_request(body, num_tries=1) -> bool:
        if not body or 'update_id' not in body:
            logging.warning('Invalid message received: {}'.format(body))
            return False

        if 'message' in body:
            message = body['message']

            chat_id = message['chat']['id']
            uid = message['from']['id'] if 'from' in message else None  # null for group chats
            if uid and not TelegramInterface.has_message_expired(message):
                logging.debug('Adding message to queue')
                accept_user_message.delay(TelegramInterface.name, uid, body, chat_id=chat_id)
                return True
            else:
                logging.warning('No sender specified, ignoring message')
                return False
        elif 'callback_query' in body:
            callback_query = body['callback_query']
            query_id = callback_query['id']
            message = callback_query['message']
            if message:
                chat_id = message['chat']['id']
                message_id = message['message_id']
                TelegramInterface.answer_callback_query(query_id, chat_id, message_id)
            else:
                logging.error('No message in callback query, probably too old, ignoring.')
            if 'message' in callback_query:
                # unfortunately, there is no way to check the age of callback itself
                chat_id = callback_query['message']['chat']['id']
                uid = message['from']['id'] if 'from' in message else None
                accept_user_message.delay(TelegramInterface.name, uid, body, chat_id=chat_id)
                return True
        else:
            logging.warning('Unknown message type')
            return False

    @staticmethod
    def parse_message(raw, num_tries=1):
        if 'message' in raw and 'text' in raw['message']:
            return parse_text_message(raw['message']['text'])
        elif 'callback_query' in raw:
            callback_query = raw['callback_query']
            data = TelegramInterface.retrieve_callback(callback_query.get('data'))
            if data:
                payload = json.loads(data)
                payload['_message_text'] = [{'value': None}]
                return {'entities': payload, 'type': 'postback'}
        return {'type': 'undefined'}

    @staticmethod
    def persist_callback(payload) -> str:
        """
        Persists payload to be used with Telegram callbacks (which can be max. 64B)
        and returns a string that can be used to retrieve the payload from redis.
        The payload will be persisted for a week.
        :param  payload     Payload to be persisted in Redis.
        :return             Unique string associated with the payload, which can be sent to Telegram.
        """
        redis = get_redis()
        key = ''.join(random.choice(string.hexdigits) for i in range(63))
        while redis.exists(key):
            key = ''.join(random.choice(string.hexdigits) for i in range(63))
        redis.set(key, payload, ex=3600 * 24 * 7)
        return key

    @staticmethod
    def retrieve_callback(key) -> Optional[str]:
        redis = get_redis()
        if redis.exists(key):
            return redis.get(key)
        return None



# -*- coding: utf-8 -*-

import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-golem',
    version='0.90-beta',
    packages=find_packages(),
    include_package_data=True,
    license='Apache Software License',
    description='Framework for creating stateful chatbots on Django.',
    long_description=README,
    url='https://github.com/prihoda/golem',
    author='David Příhoda, Jakub Drdák, Matúš Žilinec',
    author_email='david.prihoda@gmail.com',
    entry_points={
          'console_scripts': [
              'golem-admin = golem.core.admin:main',
              'golm = golm:main'
          ]
      },
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=['django', 'networkx','requests','six','sqlparse','wit==4.3.0','wheel','redis',
    'pytz','unidecode','emoji','elasticsearch','celery','python-dateutil'],
)



import logging
import os

import numpy as np
import random

import time


class GloVe:
    def __init__(self, dir='.', prefix='glove.6B'):
        """
        Loads word embeddings from disk.
        This may take a few seconds, reuse this object.
        It will take a few minutes to cache the words on first run.
        """
        self.dir = dir
        self.prefix = prefix
        self.files = [os.path.join(dir, f) for f in os.listdir(dir) if f.startswith(prefix)]
        if not self.files:  # TODO remove support for horizontally joined files and change to vertical
            raise ValueError(
                'No GloVe files found at {} with prefix {}'.format(
                    dir, prefix))
        self.offsets = {}
        self.use_weights = False
        self.weights = []
        # self.reader = None
        if not self.load_from_cache():
            self.load()
            self.create_word_cache()

    def create_word_cache(self):
        logging.warning('Creating GloVe cache at {}'.format(self.dir))
        with open(os.path.join(self.dir, 'glove.cache'), 'w') as stream:
            words = sorted(self.offsets.keys())
            for word in words:
                stream.write('{} '.format(word))
                stream.write(' '.join([str(x) for x in self.offsets[word]]))
                stream.write('\n')
            logging.warning('Wrote {} GloVe words to cache'.format(len(words)))
        with open(os.path.join(self.dir, 'glove.weights'), 'w') as stream:
            stream.write(' '.join([str(x) for x in self.weights]))
            logging.warning('Wrote {} weights to cache'.format(len(self.weights)))

    def load_from_cache(self):
        logging.info('Loading GloVe cached info')
        fpath = os.path.join(self.dir, 'glove.cache')
        if not os.path.isfile(fpath):
            logging.warning('GloVe cache does not exist')
            return False
        with open(fpath, 'r') as stream:
            for line in stream:
                tokens = line.split()
                self.offsets[tokens[0]] = [int(x) for x in tokens[1:]]
        logging.info('Loaded {} GloVe words from cache'.format(len(self.offsets)))
        if self.use_weights:
            fpath = os.path.join(self.dir, 'glove.weights')
            if not os.path.isfile(fpath):
                return False
            with open(fpath, 'r') as stream:
                weights = []
                for weight in stream.read().split():
                    weights.append(float(weight))
                self.weights = weights
        return True

    def load(self):
        logging.info('Loading GloVe')

        with open(self.files[0], 'r') as stream:
            for idx, line in enumerate(stream):
                word = line.split()[0]
                self.offsets[word] = []
        for filename in self.files:
            with open(filename) as stream:
                line, offset = stream.readline(), 0
                while line:
                    word = line.split()[0]
                    self.offsets[word].append(offset)
                    offset = stream.tell()
                    line = stream.readline()

        # try:
        #     import simstring
        #     db = simstring.writer(os.path.join(self.dir, 'simstring.db'))
        #     for word in list(self.offsets.keys()):
        #         db.insert(word)
        #     db.close()
        # except:
        #     logging.warning("Simstring not found, approximate matching will not work.")

        word_cnt = len(self.offsets)
        self.weights = self.compute_variances()
        logging.debug('Loaded {} GloVe words'.format(word_cnt))

    def get_dimension(self):
        """Returns the dimension of the output word vectors."""
        dim = 0
        for file_idx, fn in enumerate(self.files):
            with open(fn, 'r') as stream:
                line = stream.readline()
                arr = line.split()
                dim += len(arr) - 1
        return dim

    def get_vector(self, word, nearest_match=False):
        """
        Loads vector representation of a word.
        Time complexity O(1) with respect to number of words.
        :param  word
        :returns:   numpy.array on success, None otherwise.
        """
        if word not in self.offsets and nearest_match:
            word = self.get_nearest_match(word)
        if word not in self.offsets:
            return None
            #word = '########'  # FIXME use <unk>, which I don't have in the embedding right now
        if word in self.offsets:
            vec = []
            for file_idx, filename in enumerate(self.files):
                with open(filename, 'r') as stream:
                    stream.seek(self.offsets[word][file_idx])
                    line = stream.readline()
                    arr = line.split()
                    if arr[0] == word:
                        vec += [float(x) for x in arr[1:]]
            if self.use_weights:
                return np.array(vec) / self.weights
            return np.array(vec)
        return None

    def compute_variances(self):
        """Computes relative weights of all data columns."""
        variances = []
        base = 0
        for filename in self.files:
            dim_cnt = 0
            with open(filename, 'r') as f:
                for line_idx, line in enumerate(f):
                    line = line.strip()
                    if line_idx == 0:
                        dim_cnt = len(line.split(' ')) - 1
                        if dim_cnt < 1:
                            continue  # empty line
                        for i in range(dim_cnt):
                            variances.append(0.0)
                    for i, x in enumerate(line.split(' ')[1:]):
                        variances[base + i] += float(x) * float(x)
            base += dim_cnt
        return np.sqrt(np.array(variances))

    def _cos(self, u, v):
        if u is not None and v is not None:
            dot = np.dot(u, v)
            du = np.dot(u, u)
            dv = np.dot(v, v)
            return dot / np.sqrt(du * dv)
        else:
            return 0.0

    def similarity(self, w1, w2):
        """
        Computes cosine similarity between two words.
        :params:    Words w1, w2
        :returns:   Cosine similarity (0;1)
                    Returns 0.0 if word(s) not found in GloVe.
        """
        u, v = self.get_vector(w1), self.get_vector(w2)
        return self._cos(u, v)

    def extract_keywords(self, text, keywords, threshold=0.3):
        """
        Looks for words similar to keywords in text.
        :param  text        An array of words.
        :param  keywords    An array of words.
        :returns:           np.array containing tuples (word, score)
        """
        kw_vec = np.array([self.get_vector(kw) for kw in keywords])
        scalars = np.array([np.sqrt(np.dot(x, x)) for x in kw_vec])

        matches = []

        for word in text:
            u = self.get_vector(word)
            if u is None:
                continue
            dots = np.dot(kw_vec, u.T)
            u_scalar = np.sqrt(np.dot(u, u))
            score = np.mean(dots / (scalars * u_scalar))
            print(word, ':', score)
            if score > threshold:
                matches.append((word, score))
                break
        return matches

    def get_centroid(self, words):
        vec = []
        for word in words:
            v = self.get_vector(word)
            if v is not None:
                vec.append(v)
        return np.mean(vec, axis=0)

    def random_word(self):
        """Returns a randomly chosen word from this embedding."""
        return random.choice(list(self.offsets.keys()))

    # def simstring_reader(self):
    #     if self.reader != None:
    #         return self.reader
    #     elif os.path.exists(os.path.join(self.dir, 'simstring.db')):
    #         try:
    #             logging.debug("Loading simstring reader")
    #             import simstring
    #             db = simstring.reader(os.path.join(self.dir, 'simstring.db'))
    #             db.measure = simstring.cosine
    #             db.threshold = 0.2
    #             self.reader = db
    #             return db
    #         except ImportError:
    #             logging.warning("Simstring not found, nearest match not available")
    #     return None

    def get_nearest_match(self, word):
        logging.warning("Using nearest match")

        if not word:
            return None

        t = time.time()
        from golem.nlu.tools import correct
        match = correct.nearest_word(word, self)
        t = time.time() - t
        logging.warning("Nearest match: {} time: {} s".format(match, t))
        return match

    def contains(self, word):
        return self.offsets.get(word) != None



from typing import Iterable


class MenuElement:
    def __init__(self, type, title, payload=None, url=None,
                 webview_height_ratio=None, messenger_extensions=None):
        self.type = type
        self.title = title
        self.payload = payload
        self.url = url
        self.buttons = []

    def __str__(self):
        text = 'element: ' + self.title
        return text

    def add_button(self, button):
        self.buttons.append(button)
        return button


class MessageElement:
    """
    Base class for message elements.
    """
    def to_message(self, fbid):
        return {"recipient": {"id": fbid}, "message": self.to_response()}

    def to_response(self):
        pass

    def get_response_for(self, platform='fb'):
        return str(self)  # TODO

    def set_message_tag(self, tag):
        self.tag = tag

    def get_message_tag(self):
        """
        See https://developers.facebook.com/docs/messenger-platform/send-messages/message-tags
        :return: FB message tag
        """
        if hasattr(self, "tag"):
            return self.tag
        return None

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return str(self)


class SenderActionMessage(MessageElement):
    def __init__(self, action):
        self.action = action

    def to_message(self, fbid):
        return {'sender_action': self.action, 'recipient': {"id": fbid}}


class TextMessage(MessageElement):
    """
    A plain text message. You can attach buttons or quick replies.
    """

    def __init__(self, text='', buttons=None, quick_replies=None):
        from .quick_reply import QuickReply
        self.text = text
        self.buttons = buttons if buttons else []
        self.quick_replies = [QuickReply(**reply) for reply in quick_replies or []]

    def to_response(self):
        if len(self.buttons):
            return {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "button",
                        "text": self.text,
                        "buttons": [button.to_response() for button in self.buttons]
                    }
                }
            }
        response = {'text': self.text}
        if self.quick_replies:
            response["quick_replies"] = [reply.to_response() for reply in self.quick_replies]
        return response

    def __str__(self):
        text = self.text
        for button in self.buttons:
            text += "\n " + str(button)
        for reply in self.quick_replies:
            text += "\n " + str(reply)
        return text

    def add_button(self, button):
        if self.quick_replies:
            raise ValueError('Cannot add quick_replies and buttons to the same message')
        self.buttons.append(button)
        return self

    def add_reply(self, quick_reply):
        return self.add_quick_reply(quick_reply)

    def add_quick_reply(self, quick_reply):
        if self.buttons:
            raise ValueError('Cannot add quick_replies and buttons to the same message')
        elif isinstance(quick_reply, str):
            from golem.core.responses.quick_reply import QuickReply
            quick_reply = QuickReply(title=quick_reply)

        self.quick_replies.append(quick_reply)
        return self

    def create_quick_reply(self, title=None, payload=None, image_url=None):
        from .quick_reply import QuickReply
        quick_reply = QuickReply(title=title, payload=payload, image_url=image_url)
        return self.add_quick_reply(quick_reply)

    def with_replies(self, replies: Iterable):
        for reply in replies:
            self.add_quick_reply(reply)
        return self

    def with_buttons(self, buttons: Iterable):
        for button in buttons:
            self.add_button(button)
        return self


class GenericTemplateElement(MessageElement):
    """
    A horizontal card view with title, subtitle, image and buttons.
    """

    def __init__(self, title, image_url=None, subtitle=None, item_url=None, buttons=None):
        self.title = _ellipsize(str(title)) if title else None
        self.subtitle = _ellipsize(str(subtitle)) if subtitle else None
        self.image_url = image_url
        self.item_url = item_url
        self.buttons = buttons if buttons else []

    def to_response(self):
        response = {
            "title": self.title,
            "image_url": self.image_url,
            "subtitle": self.subtitle,
            "item_url": self.item_url
        }
        if self.buttons:
            response["buttons"] = [button.to_response() for button in self.buttons]
        return response

    def __str__(self):
        text = 'element: ' + self.title
        if self.subtitle:
            text += " (" + self.subtitle + ")"
        for button in self.buttons:
            text += "\n  " + str(button)
        return text

    def add_button(self, button):
        self.buttons.append(button)
        return self

class GenericTemplateMessage(MessageElement):
    """
    A horizontal list of GenericTemplateElement items.
    """

    def __init__(self, elements=None):
        self.elements = [GenericTemplateElement(**element) for element in elements or []]

    def to_response(self):
        return {
            "attachment": {
                "type": "template",
                "payload": {
                    "template_type": "generic",
                    "elements": [element.to_response() for element in self.elements[:10]]
                }
            }
        }

    def __str__(self):
        text = 'generic template:'
        for element in self.elements:
            text += "\n " + str(element)
        return text

    def add_element(self, element):
        self.elements.append(element)
        return self

    def create_element(self, **kwargs) -> GenericTemplateElement:
        element = GenericTemplateElement(**kwargs)
        self.add_element(element)
        return element


class AttachmentMessage(MessageElement):
    def __init__(self, attachment_type, url):
        self.attachment_type = attachment_type
        self.url = url

    def to_response(self):
        return {
            "attachment": {
                "type": self.attachment_type,
                "payload": {
                    "url": self.url
                }
            }
        }


# TODO stickers not yet supported by Messenger :(
# class StickerMessage(MessageElement):
#    def __init__(self, sticker_id):
#        self.sticker_id = sticker_id
#
#    def to_response(self):
#        return {
#            "sticker_id": self.sticker_id
#        }


def _ellipsize(text, max_len=80):
    if text is None:
        return None
    if len(text) >= max_len:
        return text[:max_len-5] + "..."
    return text


def _get_payload_string(payload):
    text = ''
    for entity, values in payload.items():
        if isinstance(values, list):
            for value in values:
                text += '/{}/{}/ '.format(entity, value)
            break
        text += '/{}/{}/ '.format(entity, values)
    return text



import pickle

import keras.layers as L
import numpy as np
import os
import re
from keras.callbacks import EarlyStopping
from keras.models import Sequential, load_model
from sklearn.preprocessing import LabelBinarizer
from sklearn.utils import shuffle

from golem.nlu import cleanup
from golem.nlu.nn.abs_model import NLUModel


class BowModel(NLUModel):
    def __init__(self, entity, base_dir, is_training=False):
        if not base_dir or not entity:
            raise ValueError('Base dir and entity name must be set')
        self.base_dir = base_dir
        self.entity = entity
        self.stopwords = {"i", "you", "do", "?", "!", ".", ","}

        if not is_training:
            self.load_model()

    def init_model(self):
        self.model = Sequential()
        self.model.add(L.Dense(512, input_dim=len(self.vocab)))
        self.model.add(L.Activation('relu'))
        self.model.add(L.Dense(len(self.labels) * 5))
        self.model.add(L.Activation('relu'))
        self.model.add(L.Dense(len(self.labels)))
        self.model.add(L.Activation('softmax'))
        self.model.compile(
            loss='categorical_crossentropy',  # TODO or crossentropy :-P
            optimizer='sgd',
            metrics=['accuracy']
        )

    def create_vocab(self, sentences):
        vocab = set()
        for list_ in sentences:
            for sent in list_:
                tokens = cleanup.tokenize(sent)
                vocab = vocab.union(tokens)
        vocab = vocab.difference(self.stopwords)
        vocab.add("[NUM]")
        return list(vocab)

    def load_model(self):
        with open(os.path.join(self.base_dir, self.entity, "vocab.pkl"), "rb") as f:
            self.vocab, self.labels, self.onehot = pickle.load(f)
        self.model = load_model(os.path.join(self.base_dir, self.entity, "model.h5"))
        # self.init_model(len(self.vocab))

    def encode_sentence(self, sentence):
        bow = [0.] * len(self.vocab)
        tokens = cleanup.tokenize(sentence)
        for token in tokens:
            if re.match("[0-9]+", token):
              bow[self.vocab.index("[NUM]")] = 1.0
            elif token in self.vocab:
                bow[self.vocab.index(token)] = 1.0
            else:
                print("Not in vocab")
        # print(bow)
        return bow

    def train(self, data):
        print("Training BOW model ...")
        self.labels = list(set([x['value'] for x in data]))

        if "none" not in self.labels:
            # negative samples (this is a hack)
            self.labels.insert(0, "none")

        sentences = dict((x['value'], x['samples']) for x in data)
        sentences.setdefault(None, []).append([])

        self.vocab = self.create_vocab(sentences.values())
        print("Vocabulary: ", self.vocab)

        self.init_model()
        x, y = [], []

        for value, samples in sentences.items():
            if value is None:
                value = 'none'
            for sample in samples:
                x.append(self.encode_sentence(sample))
                y.append(value)

        for _ in range(10):
            # close your eyes please
            x.append([0] * len(self.vocab))
            y.append('none')

        x, y = shuffle(x, y)

        self.onehot = LabelBinarizer()
        self.onehot.fit(self.labels)

        self.model.fit(
            np.array(x), self.onehot.transform(y),
            epochs=1000, batch_size=32,
            callbacks=[
                EarlyStopping(monitor='loss', min_delta=0.00001, verbose=1),
            ]
        )
        print("Saving model ...")
        self.model.save(os.path.join(self.base_dir, self.entity, "model.h5"))
        with open(os.path.join(self.base_dir, self.entity, "vocab.pkl"), "wb") as f:
            pickle.dump((self.vocab, self.labels, self.onehot), f)
        print("Done.")

    def predict(self, utterance, threshold=None):
        if threshold is None:
            threshold = 0.9
        label, confidence = self.get_prediction(utterance, threshold)
        if label is not None and label != 'none':
            return [{'value': label, 'confidence': confidence, 'source': 'golem'}]
        return None

    def get_prediction(self, utterance, threshold=0.9):
        x = self.encode_sentence(utterance)
        probs = self.model.predict_proba(np.array([x]), batch_size=1)[0]
        # TODO maybe confidence can be something different

        if probs.max() >= threshold:
            confidence = probs.max()
            label = self.onehot.classes_[probs.argmax()]
            return label, float(confidence)

        return None, None

    def get_probabilities(self, utterance):
        x = self.encode_sentence(utterance)
        probs = self.model.predict_proba(np.array([x]), batch_size=1)[0]
        return [(self.onehot.classes_[idx], val) for idx, val in enumerate(probs)]



from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Runs NLP tests'

    def add_arguments(self, parser):
        parser.add_argument('text', nargs='+', type=str)

    def handle(self, *args, **options):
        # from golem.nlp import train
        # train.test_all()
        raise NotImplementedError("TO DO")



#!/usr/bin/python
# Source: PUBLIC DOMAIN https://github.com/wsong/Typo-Distance/

SHIFT_COST = 3.0
INSERTION_COST = 1.0
DELETION_COST = 1.0
SUBSTITUTION_COST = 1.0

qwertyKeyboardArray = [
    ['`', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '='],
    ['q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\\'],
    ['a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\''],
    ['z', 'x', 'c', 'v', 'b', 'n', 'm', ',', '.', '/'],
    ['', '', ' ', ' ', ' ', ' ', ' ', '', '']
]

qwertyShiftedKeyboardArray = [
    ['~', '!', '@', '#', '$', '%', '^', '&', '*', '(', ')', '+'],
    ['Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I', 'O', 'P', '{', '}', '|'],
    ['A', 'S', 'D', 'F', 'G', 'H', 'J', 'K', 'L', ':', '"'],
    ['Z', 'X', 'C', 'V', 'B', 'N', 'M', '<', '>', '?'],
    ['', '', ' ', ' ', ' ', ' ', ' ', '', '']
]

layoutDict = {'QWERTY': (qwertyKeyboardArray, qwertyShiftedKeyboardArray)}

# Sets the default keyboard to use to be QWERTY.
keyboardArray = qwertyKeyboardArray
shiftedKeyboardArray = qwertyShiftedKeyboardArray


class InsertionAction:
    def __init__(self, i, c):
        self.i = i
        self.c = c

    def cost(self, s):
        return insertionCost(s, self.i, self.c)

    def perform(self, s):
        return s[:self.i] + self.c + s[self.i:]


class SubstitutionAction:
    def __init__(self, i, c):
        self.i = i
        self.c = c

    def cost(self, s):
        return substitutionCost(s, self.i, self.c)

    def perform(self, s):
        return s[:self.i] + self.c + s[(self.i + 1):]


class DeletionAction:
    def __init__(self, i):
        self.i = i

    def cost(self, s):
        return deletionCost(s, self.i)

    def perform(self, s):
        return s[:self.i] + s[(self.i + 1):]


# Returns the keyboard layout c "lives in"; for instance, if c is A, this will
# return the shifted keyboard array, but if it is a, it will return the regular
# keyboard array.  Raises a ValueError if character is in neither array
def arrayForChar(c):
    if (True in [c in r for r in keyboardArray]):
        return keyboardArray
    elif (True in [c in r for r in shiftedKeyboardArray]):
        return shiftedKeyboardArray
    else:
        raise ValueError(c + " not found in any keyboard layouts")


# Finds a 2-tuple representing c's position on the given keyboard array.  If
# the character is not in the given array, throws a ValueError
def getCharacterCoord(c, array):
    row = -1
    column = -1
    for r in array:
        if c in r:
            row = array.index(r)
            column = r.index(c)
            return (row, column)
    raise ValueError(c + " not found in given keyboard layout")


# Finds the Euclidean distance between two characters, regardless of whether
# they're shifted or not.
def euclideanKeyboardDistance(c1, c2):
    coord1 = getCharacterCoord(c1, arrayForChar(c1))
    coord2 = getCharacterCoord(c2, arrayForChar(c2))
    return ((coord1[0] - coord2[0]) ** 2 + (coord1[1] - coord2[1]) ** 2) ** (0.5)


# The cost of inserting c at position i in string s
def insertionCost(s, i, c):
    if not s or i >= len(s):
        return INSERTION_COST
    cost = INSERTION_COST
    if arrayForChar(s[i]) != arrayForChar(c):
        # We weren't holding down the shift key when we were typing the original
        # string, but started holding it down while inserting this character, or
        # vice versa.  Either way, this action should have a higher cost.
        cost += SHIFT_COST
    cost += euclideanKeyboardDistance(s[i], c)
    return cost


# The cost of omitting the character at position i in string s
def deletionCost(s, i):
    return DELETION_COST


# The cost of substituting c at position i in string s
def substitutionCost(s, i, c):
    cost = SUBSTITUTION_COST
    if len(s) == 0 or i >= len(s):
        return INSERTION_COST
    if arrayForChar(s[i]) != arrayForChar(c):
        # We weren't holding down the shift key when we were typing the original
        # string, but started holding it down while inserting this character, or
        # vice versa.  Either way, this action should have a higher cost.
        cost += SHIFT_COST
    cost += euclideanKeyboardDistance(s[i], c)
    return cost


# Finds the typo distance (a floating point number) between two strings, based
# on the canonical Levenshtein distance algorithm.
def typoDistance(s, t, layout='QWERTY'):
    if layout in layoutDict:
        keyboardArray, shiftedKeyboardArray = layoutDict[layout]
    else:
        raise KeyError(layout + " keyboard layout not supported")

    # A multidimensional array of 0s with len(s) rows and len(t) columns.
    d = [[0] * (len(t) + 1) for i in range(len(s) + 1)]

    for i in range(len(s) + 1):
        d[i][0] = sum([deletionCost(s, j - 1) for j in range(i)])
    for i in range(len(t) + 1):
        intermediateString = ""
        cost = 0.0
        for j in range(i):
            cost += insertionCost(intermediateString, j - 1, t[j - 1])
            intermediateString = intermediateString + t[j - 1]
        d[0][i] = cost

    for j in range(1, len(t) + 1):
        for i in range(1, len(s) + 1):
            if s[i - 1] == t[j - 1]:
                d[i][j] = d[i - 1][j - 1]
            else:
                delCost = deletionCost(s, i - 1)
                insertCost = insertionCost(s, i, t[j - 1])
                subCost = substitutionCost(s, i - 1, t[j - 1])
                d[i][j] = min(d[i - 1][j] + delCost,
                              d[i][j - 1] + insertCost,
                              d[i - 1][j - 1] + subCost)

    return d[len(s)][len(t)]


# Returns a list of the possible actions than can be performed on a string s.
def getPossibleActions(s, layout='QWERTY'):
    if layout in layoutDict:
        keyboardArray, shiftedKeyboardArray = layoutDict[layout]
    else:
        raise KeyError(layout + " keyboard layout not supported")
    actions = []
    for i in range(len(s)):
        actions.append(DeletionAction(i))
        for key in (sum([r for r in keyboardArray], []) + sum([r for r in shiftedKeyboardArray], [])):
            actions.append(SubstitutionAction(i, key))
            actions.append(InsertionAction(i, key))
    return actions


# Returns a generator which generates all possible typos that are less than
# or equal to the given maximum typo distance d from the start phrase s.  Based
# on Knuth's Algorithm F in Pre-Facsicle 3A.
def typoGenerator(s, d, layout='QWERTY'):
    t = 0
    r = d
    actions = getPossibleActions(s, layout)
    # A list of the actions we will perform on this string.  We store the
    # indices of the actions in the action list above; the 0th action in
    # currentActions is simply a placeholder value and is not used.
    c = [len(actions)]
    # We update the string as we go along, so that we can check the cost
    # of performing, say, an individual assertion.  This variable is "one
    # behind" the actions in c, so we can check the cost of performing
    # the last action in c versus some other action.
    changedString = s

    while (True):
        if t == 0:
            # No actions
            yield s
        else:
            # Perform the last action
            yield actions[c[t]].perform(changedString)

        # Let's try adding a new action
        if c[t] > 0 and r >= actions[0].cost(changedString):
            t += 1
            c.append(0)
            r -= actions[0].cost(changedString)
            changedString = s
            for a in c[1:-1]:
                changedString = actions[a].perform(changedString)
            continue

        while True:
            if t == 0:
                return
            i = 1
            brokeOut = False
            # Let's try replacing the last action with a new candidate
            while (c[t - 1] > c[t] + i):
                if r >= (actions[c[t] + i].cost(changedString) - actions[c[t]].cost(changedString)):
                    # Our new candidate is cheap enough; use it.
                    c[t] += i
                    r -= actions[c[t]].cost(changedString) - actions[c[t] - i].cost(changedString)
                    changedString = s
                    for a in c[1:-1]:
                        changedString = actions[a].perform(changedString)
                    brokeOut = True
                    break
                else:
                    # The candidate was too expensive; move onto the next
                    i += 1

            if not brokeOut:
                # Let's try removing an action
                r += actions[c[t]].cost(changedString)
                c.pop(t)
                changedString = s
                for a in c[1:-1]:
                    changedString = actions[a].perform(changedString)
                t -= 1
            else:
                break



from django.conf.urls import url

from .views import *

urlpatterns = [
    # url(r'(?P<uid>[0-9]+)', view=webgui, name='webgui'),
    url(r'^$', view=webgui, name='webgui'),
    url(r'^login/?$', view=do_login, name='do_login'),
    url(r'^logout/?$', view=do_logout, name='do_logout'),
    url(r'^last_change/$', view=get_last_change, name='last_change')
]



import json
import logging
import pickle
from datetime import datetime

import numpy as np
import os
import tensorflow as tf
from django.conf import settings

from golem.nlu import utils, cleanup
from golem.nlu.nn.abs_model import NLUModel

MAX_WORD_COUNT = settings.NLP_CONFIG.get("MAX_WORD_COUNT", 10)


class RecurrentIntentModel(NLUModel):
    def __init__(self, entity, base_dir, is_training=False):
        if not base_dir or not entity:
            raise ValueError('Base dir and entity name must be set')
        self.base_dir = base_dir
        self.entity = entity
        self.graph = tf.Graph()
        self.session = tf.Session(graph=self.graph)  # , config=tf.ConfigProto(log_device_placement=True))
        self.file_writer = tf.summary.FileWriter(os.path.join(base_dir, entity, "tensorboard", str(datetime.now())))
        self.loaded = False
        self.label_cnt = None

        if not is_training:
            self.glove = utils.get_glove()
            with open(os.path.join(base_dir, 'metadata.pkl'), 'rb') as f:
                self.metadata = pickle.load(f)
                self.labels = self.metadata['labels']

    def destroy(self):
        self.session.close()

    def make_weights(self, shape, name=None):
        return tf.Variable(tf.truncated_normal(shape, stddev=0.1), name=name)

    def make_biases(self, shape, name=None):
        return tf.Variable(tf.constant(0.1, shape=shape), name=name)

    def loss_fn(self):
        loss = tf.reduce_mean(
            tf.nn.softmax_cross_entropy_with_logits(
                labels=self.y, logits=self.y_
            )
        )
        return loss

    def get_last_component(self, lstm_output, length):
        batch_size = tf.shape(lstm_output)[0]
        max_length = tf.shape(lstm_output)[1]
        out_size = int(lstm_output.get_shape()[2])
        index = tf.range(0, batch_size) * max_length + (length - 1)
        flat = tf.reshape(lstm_output, [-1, out_size])
        relevant = tf.gather(flat, index)
        return relevant

    def init_network(self, feature_cnt, label_cnt, max_word_cnt, batch_size=1):
        memory_size = int(feature_cnt)

        self.x = tf.placeholder(tf.float32, shape=[None, max_word_cnt, feature_cnt], name='x')
        self.y = tf.placeholder(tf.float32, shape=[None, label_cnt], name='y')
        self.keep_prob = tf.placeholder_with_default(tf.cast(1.0, tf.float32), [], "keep_prob")
        self.lengths = tf.placeholder_with_default(tf.fill([batch_size], max_word_cnt), [batch_size], "lengths")

        with tf.name_scope("rnn"):

            fwd_lstm = tf.nn.rnn_cell.DropoutWrapper(
                tf.contrib.rnn.LSTMCell(memory_size),
                input_keep_prob=self.keep_prob,
                output_keep_prob=self.keep_prob,
                state_keep_prob=self.keep_prob
            )
            bwd_lstm = tf.nn.rnn_cell.DropoutWrapper(
                tf.contrib.rnn.LSTMCell(memory_size),
                input_keep_prob=self.keep_prob,
                output_keep_prob=self.keep_prob,
                state_keep_prob=self.keep_prob
            )

            rnn_1, state = tf.nn.bidirectional_dynamic_rnn(
                cell_fw=fwd_lstm,
                cell_bw=bwd_lstm,
                inputs=self.x,
                sequence_length=self.lengths,
                dtype=tf.float32
            )

            foo = self.get_last_component(rnn_1[0], self.lengths)
            bar = self.get_last_component(rnn_1[1], self.lengths)

        flat = tf.reshape(tf.stack([foo, bar], axis=2), [-1, memory_size * 2])

        dense_1 = tf.layers.dense(flat, label_cnt * 8, tf.nn.leaky_relu, name="dense_1")
        drop = tf.layers.dropout(dense_1, self.keep_prob)
        self.y_ = tf.layers.dense(drop, label_cnt, name="classifier")
        self.preds = tf.nn.softmax(self.y_)

    def train(self, batcher, num_iterations, keep_prob):

        with self.graph.as_default():

            feature_cnt = batcher.dim
            label_cnt = len(batcher.labels)
            max_word_cnt = batcher.max_words

            batch_size = settings.NLP_CONFIG.get("BATCH_SIZE", 32)
            initial_step = settings.NLP_CONFIG.get("INITIAL_STEP", 1e-3)
            use_tensorboard = settings.NLP_CONFIG.get("USE_TENSORBOARD", False)

            self.init_network(feature_cnt, label_cnt, max_word_cnt, batch_size=batch_size)
            loss = self.loss_fn()
            optimizer = tf.train.AdamOptimizer(initial_step).minimize(loss)
            accuracy = tf.reduce_mean(tf.metrics.accuracy(tf.argmax(self.preds, 1), tf.argmax(self.y, 1)))

            self.session.run(tf.global_variables_initializer())
            self.session.run(tf.local_variables_initializer())

            self.file_writer.add_graph(self.graph)
            tf.summary.scalar("loss", loss)
            tf.summary.scalar("accuracy", accuracy)
            tensorboard_op = tf.summary.merge_all() if use_tensorboard else None

            past_losses = [1.] * 25

            for i in range(num_iterations):

                # start_idx = random.randint(0, sample_cnt - batch_size - 1)
                # end_idx = start_idx + batch_size

                # _x_ = features[start_idx:end_idx]
                # _y_ = labels_onehot[start_idx:end_idx]
                # _junk_ = junk[start_idx:end_idx]

                # if i % 100 == 0:  TODO should this stay ?!
                #     total_junk_idx = random.randint(0, batch_size - 1)
                #     _x_[total_junk_idx] = [np.random.random(feature_cnt) for j in range(word_cnt)]
                #     _y_[total_junk_idx] = np.zeros(label_cnt)

                # trashify TODO already done in batcher
                # for j in range(len(_x_)):
                #     junks = _junk_[j]
                #     for junk_pos in junks:
                #         _x_[j][junk_pos] = np.random.random(feature_cnt)
                # print(_x_)

                # zipped = list(zip(_x_, _y_))
                # random.shuffle(zipped)
                # _x_, _y_ = zip(*zipped)

                _x_, _y_ = batcher.next_batch(batch_size)

                optimizer.run(feed_dict={self.x: _x_, self.y: _y_, self.keep_prob: keep_prob},
                              session=self.session)
                if use_tensorboard:
                    tensorboard_result = self.session.run(tensorboard_op, feed_dict={self.x: _x_, self.y: _y_})
                    self.file_writer.add_summary(tensorboard_result, i)

                xe = self.session.run(loss, feed_dict={self.x: _x_, self.y: _y_})

                print('Iteration ', i, '/', num_iterations,
                      'Xentropy:', xe,
                      'Accuracy:', self.session.run(accuracy, feed_dict={self.x: _x_, self.y: _y_}))

                past_losses.insert(0, xe)
                past_losses = past_losses[:25]
                if np.max(past_losses) < 0.05:
                    print("Early stopping, loss < 0.05")
                    break

            # save model
            saver = tf.train.Saver()
            saver.save(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))
            with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json'), 'w') as f:
                json.dump({'feature_cnt': feature_cnt, 'label_cnt': label_cnt, 'word_cnt': max_word_cnt}, f)

    def load(self):
        with open(os.path.join(self.base_dir, self.entity, 'tf_metadata.json')) as f:
            metadata = json.load(f)
        feature_cnt, label_cnt = metadata['feature_cnt'], metadata['label_cnt']
        word_cnt = metadata['word_cnt']
        print('features:', feature_cnt, 'labels:', label_cnt)
        print('Restoring from {}'.format(self.base_dir))
        self.init_network(feature_cnt, label_cnt, word_cnt)
        self.label_cnt = label_cnt
        saver = tf.train.Saver()
        saver.restore(self.session, os.path.join(self.base_dir, self.entity, 'tf_session'))

    def predict(self, utterance: str, threshold=None):

        if threshold is None:
            threshold = 0.7  # TODO

        vector = self.word2vec(utterance)
        y_pred = self.get_predictions(vector)[0]
        logging.debug(y_pred)
        print(y_pred)
        y_pred = [[i, score] for i, score in enumerate(y_pred) if score > threshold]
        y_pred.sort(key=lambda x: x[1], reverse=True)

        if len(y_pred) > 0:
            value = self.labels[y_pred[0][0]]
            prob = y_pred[0][1]
            print('> Predicted {}: {} Prob: {}%'.format(self.entity, value, prob * 100))
            print('All scores > threshold are:', y_pred)
            if value is not None and value != 'none':
                return [{'value': value, 'confidence': float(prob)}]
        return None

    def word2vec(self, text):
        """
        Converts text to a vector.
        :param text     Text to be vectorized.
        :returns        A numpy array.
        """
        features = [np.zeros(self.glove.get_dimension()) for x in range(10)]
        tokens = cleanup.tokenize(text, stemming=False, language="cz")  # FIXME stemming, language
        imputer_rules = utils.get_imputation_rules(self.entity)  # TODO
        clean_text = cleanup.imputer(tokens, imputer_rules)
        print("Tokens:", clean_text)
        for idx, word in enumerate(clean_text):
            if idx > MAX_WORD_COUNT:
                logging.warning('Message too long!')  # FIXME allow longer messages
            vec = self.glove.get_vector(word)
            if vec is not None:
                features[idx] = vec
        return np.array([features])

    def get_predictions(self, features):
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            return self.session.run(self.preds, feed_dict={self.x: features})

    def test(self, x, y):
        with self.graph.as_default():
            if not self.loaded:
                self.load()
                self.loaded = True

            # accuracy = tf.metrics.accuracy(predictions=tf.argmax(self.y_, 1), labels=tf.argmax(self.y, 1))
            scores, predictions, losses = [], [], []
            output = []
            loss_fn = self.loss_fn()
            for _x_, _y_ in zip(x, y):
                labels = np.zeros(self.label_cnt)
                if _y_ != -1:
                    labels[_y_] = 1.0

                y_ = self.session.run(self.preds, feed_dict={self.x: _x_})
                loss = self.session.run(loss_fn, feed_dict={self.x: _x_, self.y: [labels]})
                arg = np.argmax(y_)
                predictions.append(arg)
                losses.append(loss)
                maxscore = np.max(y_)
                mean = np.mean(y_)
                if (arg == _y_ and maxscore > 0.8) or (maxscore < 0.8 and _y_ == -1):
                    print("Correct {}".format(maxscore))
                    scores.append(maxscore)
                    output.append((True, maxscore, arg))
                else:
                    print("Incorrect {}".format(maxscore))
                    scores.append(0.0)
                    output.append((False, maxscore, arg))
            correct = sorted(scores, reverse=True).index(0.0) if 0.0 in scores else len(scores)
            print("[{}] Accuracy score: {}% or {}/{}".format(self.entity, np.mean(scores) * 100,
                                                             correct, len(scores)))
            print("[{}] Mean Cross Entropy Loss: {}".format(self.entity, np.mean(losses)))
            return output, predictions



import dateutil.parser

def json_deserialize(obj):
    #print('Deserializing:', obj)
    if obj.get('__type__') == 'datetime':
        return dateutil.parser.parse(obj.get('value'))
    return obj

def json_serialize(obj):
    from datetime import datetime
    if isinstance(obj, datetime):
        return {'__type__':'datetime', 'value': obj.isoformat()}
    raise TypeError ("Error saving entity value. Type %s not serializable: %s" % (type(obj), obj))



import datetime
import logging

import requests
from django.conf import settings

from golem.core.chat_session import ChatSession
from golem.core.message_parser import parse_text_message
from golem.core.persistence import get_redis
from golem.core.responses.buttons import *
from golem.core.responses.quick_reply import QuickReply
from golem.core.responses.responses import *
from golem.core.responses.settings import ThreadSetting, GreetingSetting, GetStartedSetting, MenuSetting
from golem.core.responses.templates import ListTemplate
from golem.core.serialize import json_deserialize
from golem.tasks import accept_user_message


class FacebookInterface():
    name = 'facebook'
    prefix = 'fb'
    TEXT_LENGTH_LIMIT = 320

    # Post function to handle Facebook messages
    @staticmethod
    def accept_request(request):
        # Facebook recommends going through every entry since they might send
        # multiple messages in a single call during high load.
        for entry in request['entry']:
            for raw_message in entry['messaging']:
                ts_datetime = datetime.datetime.fromtimestamp(int(raw_message['timestamp']) / 1000)
                crr_datetime = datetime.datetime.utcnow()
                diff = crr_datetime - ts_datetime
                if diff.total_seconds() < settings.GOLEM_CONFIG.get('MSG_LIMIT_SECONDS', 15):
                    # get and persist user and page ids
                    print('INCOMING RAW FB MESSAGE: {}'.format(raw_message))
                    user_id = raw_message['sender']['id']
                    page_id = entry['id']
                    chat_id = FacebookInterface.create_chat_id(page_id, user_id)
                    meta = {"user_id": user_id, "page_id": page_id}
                    session = ChatSession(FacebookInterface, chat_id, meta=meta)
                    FacebookInterface.fill_session_profile(session)
                    # Confirm accepted message
                    FacebookInterface.post_message(session, SenderActionMessage('mark_seen'))
                    # Add it to the message queue
                    accept_user_message.delay(session.to_json(), raw_message)
                elif raw_message.get('timestamp'):
                    print("Delay {} too big, ignoring message!".format(diff))
                    print(raw_message)

    @staticmethod
    def chat_id_to_page_id(chat_id):
        return chat_id.split('_', maxsplit=1)[0]

    @staticmethod
    def create_chat_id(page_id, fbid):
        return "{}_{}".format(page_id, fbid)

    @staticmethod
    def get_page_token(page_id):
        if 'FB_PAGE_TOKENS' in settings.GOLEM_CONFIG:
            tokens = settings.GOLEM_CONFIG.get('FB_PAGE_TOKENS')
            if page_id not in tokens:
                raise Exception('Page id "{}" not in tokens: {}'.format(page_id, tokens))
            return tokens.get(page_id)
        elif 'FB_PAGE_TOKEN' in settings.GOLEM_CONFIG:  # there is just one page
            return settings.GOLEM_CONFIG.get("FB_PAGE_TOKEN")
        return None

    @staticmethod
    def load_profile(user_id, page_id, cache=True):

        db = get_redis()
        key = 'fb_profile_' + user_id

        if not cache or not db.exists(key):
            print('Loading fb profile...')

            url = "https://graph.facebook.com/v2.6/" + user_id
            params = {
                'fields': 'first_name,last_name,profile_pic,locale,timezone,gender',
                'access_token': FacebookInterface.get_page_token(page_id)
            }
            res = requests.get(url, params=params)
            if not res.status_code == requests.codes.ok:
                print("!!!!!!!!!!!!!!!!!!!!!!! ERROR load_profile:")
                print(res)
                return {}

            db.set(key, json.dumps(res.json()), ex=3600 * 24 * 14)  # save value, expire in 14 days

        return json.loads(db.get(key).decode('utf-8'))

    @staticmethod
    def fill_session_profile(session: ChatSession):
        if not session:
            raise ValueError("Session is None")
        user_id, page_id = session.meta.get("user_id"), session.meta.get("page_id")
        profile_dict = FacebookInterface.load_profile(user_id, page_id)
        session.profile.first_name = profile_dict.get("first_name")
        session.profile.last_name = profile_dict.get("last_name")
        return session

    @staticmethod
    def post_message(session: ChatSession, response):
        fbid = session.meta.get("user_id")
        page_id = session.meta.get("page_id")

        if isinstance(response, SenderActionMessage):
            request_mode = "messages"
            response_dict = {
                'sender_action': response.action,
                'recipient': {"id": fbid},
            }
        elif isinstance(response, MessageElement):
            message_tag = response.get_message_tag()
            message = FacebookInterface.to_message(response)
            response_dict = {
                "recipient": {"id": fbid},
                "message": message,
                "messaging_type": "MESSAGE_TAG" if message_tag else "RESPONSE",
                "tag": message_tag,
            }
            request_mode = "messages"
        else:
            raise ValueError('Error: Invalid message type: {}: {}'.format(type(response), response))

        FacebookInterface._do_post(request_mode, response_dict, page_id)

    @staticmethod
    def post_setting(page_id, response):
        if isinstance(response, ThreadSetting):
            request_mode = "thread_settings"
            response_dict = FacebookInterface.to_setting(response)
            print('SENDING SETTING:', response_dict)
            FacebookInterface._do_post(request_mode, response_dict, page_id)
        else:
            raise ValueError('Error: Invalid message type: {}: {}'.format(type(response), response))

    @staticmethod
    def _do_post(request_mode, response_dict, page_id):
        prefix_post_message_url = 'https://graph.facebook.com/v2.6/me/'
        token = FacebookInterface.get_page_token(page_id)
        post_message_url = prefix_post_message_url + request_mode + '?access_token=' + token

        r = requests.post(post_message_url,
                          headers={"Content-Type": "application/json"},
                          data=json.dumps(response_dict, default=json_serialize))
        if r.status_code != 200:
            logging.error('ERROR: MESSAGE REFUSED: {}'.format(response_dict))
            logging.error('ERROR: {}'.format(r.text))
            logging.exception(r.json()['error']['message'])

    @staticmethod
    def to_setting(response):
        if isinstance(response, GreetingSetting):
            return {
                "greeting": {'text': response.message},
                "setting_type": "greeting"
            }
        elif isinstance(response, GetStartedSetting):
            return {
                "call_to_actions": [{'payload': json.dumps(response.payload, default=json_serialize)}],
                "setting_type": "call_to_actions",
                "thread_state": "new_thread"
            }
        elif isinstance(response, MenuSetting):
            return {
                "call_to_actions": [FacebookInterface.to_setting(element) for element in response.elements[:10]],
                "setting_type": "call_to_actions",
                "thread_state": "existing_thread"
            }
        elif isinstance(response, MenuElement):
            r = {
                "title": response.title,
                "type": response.type,
            }
            if response.payload:
                r['payload'] = json.dumps(response.payload, default=json_serialize)
            if response.url:
                r['url'] = response.url
            return r
        raise ValueError('Error: Invalid setting type: {}: {}'.format(type(response), response))

    @staticmethod
    def to_message(response):
        if isinstance(response, TextMessage):
            if response.buttons:
                return {
                    "attachment": {
                        "type": "template",
                        "payload": {
                            "template_type": "button",
                            "text": response.text[:FacebookInterface.TEXT_LENGTH_LIMIT],
                            "buttons": [FacebookInterface.to_message(button) for button in response.buttons]
                        }
                    }
                }
            message = {'text': response.text[:FacebookInterface.TEXT_LENGTH_LIMIT]}
            if response.quick_replies:
                message["quick_replies"] = [FacebookInterface.to_message(reply) for reply in response.quick_replies]
            return message

        elif isinstance(response, GenericTemplateMessage):
            return {
                "attachment": {
                    "type": "template",
                    "payload": {
                        "template_type": "generic",
                        "elements": [FacebookInterface.to_message(element) for element in response.elements[:10]]
                    }
                }
            }

        elif isinstance(response, AttachmentMessage):
            return {
                "attachment": {
                    "type": response.attachment_type,
                    "payload": {
                        "url": response.url
                    }
                }
            }

        elif isinstance(response, GenericTemplateElement):
            message = {
                "title": response.title,
                "image_url": response.image_url,
                "subtitle": response.subtitle,
                "item_url": response.item_url
            }
            if response.buttons:
                message["buttons"] = [FacebookInterface.to_message(button) for button in response.buttons]
            return message

        elif isinstance(response, QuickReply):
            return response.to_response()

        elif isinstance(response, Button):
            return response.to_response()

        elif isinstance(response, ListTemplate):
            return response.to_response()

        raise ValueError('Error: Invalid message type: {}: {}'.format(type(response), response))

    @staticmethod
    def send_settings(setting_list):
        for setting in setting_list:
            if 'FB_PAGE_TOKENS' in settings.GOLEM_CONFIG:
                for page_id in settings.GOLEM_CONFIG.get('FB_PAGE_TOKENS'):
                    FacebookInterface.post_setting(page_id, setting)
            elif 'FB_PAGE_TOKEN' in settings.GOLEM_CONFIG:
                FacebookInterface.post_setting("", setting)

    @staticmethod
    def processing_start(session: ChatSession):
        # Show typing animation
        FacebookInterface.post_message(session, SenderActionMessage('typing_on'))

    @staticmethod
    def processing_end(session: ChatSession):
        pass

    @staticmethod
    def state_change(state):
        pass

    @staticmethod
    def parse_message(raw_message, num_tries=1):
        if 'postback' in raw_message:
            payload = json.loads(raw_message['postback']['payload'], object_hook=json_deserialize)
            payload['_message_text'] = [{'value': None}]
            return {'entities': payload, 'type': 'postback'}
        elif 'message' in raw_message:
            if 'sticker_id' in raw_message['message']:
                return FacebookInterface.parse_sticker(raw_message['message']['sticker_id'])
            if 'attachments' in raw_message['message']:
                attachments = raw_message['message']['attachments']
                return FacebookInterface.parse_attachments(attachments)
            if 'quick_reply' in raw_message['message']:
                payload = json.loads(raw_message['message']['quick_reply'].get('payload'), object_hook=json_deserialize)
                if payload:
                    payload['_message_text'] = [{'value': raw_message['message']['text']}]
                    return {'entities': payload, 'type': 'postback'}
            if 'text' in raw_message['message']:
                return parse_text_message(raw_message['message']['text'])
        return {'type': 'undefined'}

    @staticmethod
    def parse_sticker(sticker_id):
        if sticker_id in [369239383222810, 369239343222814, 369239263222822]:
            return {'entities': {'emoji': 'thumbs_up_sign', '_message_text': None}, 'type': 'message'}

        return {'entities': {'sticker_id': sticker_id, '_message_text': None}, 'type': 'message'}

    @staticmethod
    def parse_attachments(attachments):
        entities = {
            'intent': [],
            'current_location': [],
            'attachment': [],
            '_message_text': [{'value': None}]
        }
        for attachment in attachments:
            if 'coordinates' in attachment['payload']:
                coordinates = attachment['payload']['coordinates']
                entities['current_location'].append({'value': attachment['title'], 'name': attachment['title'],
                                                     'coordinates': coordinates, 'timestamp': datetime.datetime.now()})
            if 'url' in attachment['payload']:
                url = attachment['payload']['url']
                # TODO: add attachment type by extension
                entities['attachment'].append({'value': url})
                entities['intent'].append({'value': 'attachment'})
        return {'entities': entities, 'type': 'message'}



import logging

import requests

from golem.core.logging.abs_logger import MessageLogger


class ChatbaseLogger(MessageLogger):
    def __init__(self, api_key):
        super().__init__()
        self.base_url = 'https://chatbase.com/api'
        self.api_key = api_key
        if self.api_key is None:
            logging.warning("Chatbase API key not provided, will not log!")

    def _interface_to_platform(self, interface: str):
        if interface is None:
            return None
        return interface  # TODO

    def log_user_message(self, dialog, accepted_time, state, message: dict, type, entities):
        payload = {
            "api_key": self.api_key,
            "type": "user",
            "user_id": dialog.session.chat_id,
            "time_stamp": int(accepted_time * 1000),
            "platform": self._interface_to_platform(dialog.session.interface.name),
            "message": str(message),
            "intent": dialog.context.get("intent", max_age=0),
            "session_id": state,
            "not_handled": False,  # TODO
            "version": "1.0",  # TODO
        }
        response = requests.post(self.base_url + "/message", params=payload)
        if not response.ok:
            logging.error("Chatbase request with code %d, reason: %s", response.status_code, response.reason)
        return response.ok

    def log_bot_message(self, dialog, accepted_time, state, message):
        payload = {
            "api_key": self.api_key,
            "type": "agent",
            "user_id": dialog.session.chat_id,
            "time_stamp": int(accepted_time * 1000),
            "platform": self._interface_to_platform(dialog.session.interface.name),
            "message": str(message),
            "intent": dialog.context.get("intent", max_age=0),
            "session_id": state,
            "not_handled": False,  # TODO
            "version": "1.0",  # TODO
        }
        response = requests.post(self.base_url + "/message", params=payload)
        if not response.ok:
            logging.error("Chatbase request with code %d, reason: %s", response.status_code, response.reason)
        return response.ok
